



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
python -c "import pandas; pandas.test(extra_args=['-m not clipboard and not single_cpu and not slow and not network and not db', '-n=2', '-k', 'not (_not_a_real_test or test_floordiv_axis0_numexpr_path[python-pow] or TestDatetime64ArrayLikeComparisons or test_apply_out_of_range or TestNonNano)', '--no-strict-data-files'])"
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
